use eyre::eyre;
use serde::{Deserialize, Serialize};
use sqlx::SqlitePool;
use std::collections::HashMap;
use tokio::fs;
use tokio::sync::mpsc;

#[derive(Deserialize, Debug, Serialize, Copy, Clone)]
#[serde(rename_all = "lowercase")]
enum MonsterProperties {
    Fiery,
    Dragon,
    Undead,
    Spectral,
    Demon,
    Leafy,
    Kalphite,
    Vampyre,
    Shade,
    Golem,
    Penance,
    Xerician,
}

#[derive(Debug, Clone)]
struct ValidMonster {
    id: i32,
    name: String,
    health: i32,
    attributes: Vec<MonsterProperties>,
    attack_level: i32,
    strength_level: i32,
    defence_level: i32,
    ranged_level: i32,
    magic_level: i32,
    defence_stab: i32,
    defence_slash: i32,
    defence_crush: i32,
    defence_magic: i32,
    defence_ranged: i32,
}

impl TryFrom<Monster> for ValidMonster {
    type Error = eyre::Error;

    fn try_from(value: Monster) -> Result<Self, Self::Error> {
        Ok(ValidMonster {
            id: value.id,
            name: value.name.clone(),
            health: value
                .health
                .ok_or(eyre!("Couldn't unwrap health for {}", value.name))?,
            attributes: value.attributes,
            attack_level: value.attack_level,
            strength_level: value.strength_level,
            defence_level: value.defence_level,
            ranged_level: value.ranged_level,
            magic_level: value.magic_level,
            defence_stab: value.defence_stab,
            defence_slash: value.defence_slash,
            defence_crush: value.defence_crush,
            defence_magic: value.defence_magic,
            defence_ranged: value.defence_ranged,
        })
    }
}

#[derive(Deserialize, Debug)]
struct Monster {
    id: i32,
    name: String,
    #[serde(rename = "hitpoints")]
    health: Option<i32>,
    attributes: Vec<MonsterProperties>,
    attack_level: i32,
    strength_level: i32,
    defence_level: i32,
    ranged_level: i32,
    magic_level: i32,
    defence_stab: i32,
    defence_slash: i32,
    defence_crush: i32,
    defence_magic: i32,
    defence_ranged: i32,
}

#[tokio::main]
async fn main() -> eyre::Result<()> {
    let pool = SqlitePool::connect("sqlite:monsters.db").await?;

    let f = fs::read_to_string("monsters-complete.json").await?;
    let a = serde_json::from_str::<HashMap<usize, Monster>>(&f)?;

    let validmonsters: Vec<ValidMonster> = a
        .into_iter()
        .filter_map(|(_, monster)| monster.try_into().ok())
        .collect();

    println!("Found {} valid monsters", validmonsters.len());

    let (tx, mut rx) = mpsc::unbounded_channel();
    validmonsters.into_iter().for_each(|monster| {
        let sender = tx.clone();
        match sender.send(monster.clone()) {
            Ok(_) => {
                println!("Sent {} successfully", &monster.name)
            }
            Err(_) => {
                eprintln!(
                    "{}",
                    eyre::eyre!("Couldn't send {} successfully", &monster.name)
                )
            }
        };
    });

    drop(tx);

    while let Some(monster) = rx.recv().await {
        println!("Writing {}", &monster.name);
        add_monster(&pool, &monster).await?;
    }

    pool.close().await;
    Ok(())
}

async fn add_monster(pool: &SqlitePool, monster: &ValidMonster) -> eyre::Result<()> {
    let mut conn = pool.acquire().await?;
    // Insert the task, then obtain the ID of this row

    let attributes = serde_json::to_string(&monster.attributes).unwrap();

    sqlx::query!(
        r#"
INSERT INTO monsters ( id, name, health, attributes, attack_level, strength_level, defence_level, ranged_level, magic_level, defence_stab, defence_slash, defence_crush, defence_magic, defence_ranged )
VALUES ( ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12 , ?13, ?14 )
ON CONFLICT(id)
DO UPDATE SET
    name=excluded.name,
    health=excluded.health,
    attributes=excluded.attributes,
    attack_level=excluded.attack_level,
    strength_level=excluded.strength_level,
    defence_level=excluded.defence_level,
    ranged_level=excluded.ranged_level,
    magic_level=excluded.magic_level,
    defence_stab=excluded.defence_stab,
    defence_slash=excluded.defence_slash,
    defence_crush=excluded.defence_crush,
    defence_magic=excluded.defence_magic,
    defence_ranged=excluded.defence_ranged
        "#,
        monster.id,
        monster.name,
        monster.health,
        attributes,
        monster.attack_level,
        monster.strength_level,
        monster.defence_level,
        monster.ranged_level,
        monster.magic_level,
        monster.defence_stab,
        monster.defence_slash,
        monster.defence_crush,
        monster.defence_magic,
        monster.defence_ranged
    )
        .execute(&mut conn)
        .await?;

    Ok(())
}
